# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['apres']

package_data = \
{'': ['*']}

install_requires = \
['matplotlib>=3.4.3,<4.0.0', 'netCDF4>=1.5.2,<2.0.0', 'numpy>=1.17.2,<2.0.0']

extras_require = \
{'remote': ['fsspec>=2024.10.0,<2025.0.0']}

entry_points = \
{'console_scripts': ['apres_to_nc = apres.apres_to_nc:main',
                     'nc_to_apres = apres.nc_to_apres:main',
                     'plot_apres = apres.plot_apres:main',
                     'read_apres = apres.read_apres:main',
                     'write_apres = apres.write_apres:main']}

setup_kwargs = {
    'name': 'bas-apres',
    'version': '0.3.0',
    'description': 'Package for working with BAS ApRES files',
    'long_description': '# apres\n\nThis project enables ApRES .dat files to be read, rewritten, and converted.\n\n## The apres package\n\nThe apres package contains a number of classes for working with ApRES .dat files.\n\n* The `ApRESBurst` class manages reading/writing a single burst block (the textual header and the binary data) from/to an opened ApRES .dat file.\n* The `ApRESFile` class is a context manager for ApRES .dat files.  The class wraps the `ApRESBurst` class and can read a variable number of burst blocks from a file (i.e. a single burst file or a timeseries file).  It also contains methods for subsetting an ApRES file, and for converting the file to netCDF4.\n\n### Loading data from remote cloud storage\n\nAs of version 0.3.0, the apres package has the ability to load data held on remote cloud storage platforms such as those offered by Google or Amazon.  To take advantage of this, the optional `remote` package extra must be installed as well as the specific [fsspec](https://filesystem-spec.readthedocs.io/en/stable/index.html) package for the required cloud storage platforms (e.g. `gcsfs` or `s3fs`).  See [Install remote extra](#install-remote-extra) for installation details.\n\nTo specify a remote ApRES file, the `path` passed to `ApRESFile()` should be a valid URI for the given cloud storage platform, e.g. `s3://<bucket_name>/<object_path>`.  In addition, any filesystem-specific `fsspec` options required for accessing the file (e.g. credentials) should be specified in the `fs_opts` `dict`, passed as a *kwarg* to `ApRESFile()`.  For example:\n\n```python\n    with ApRESFile(\'s3://bucket/data.dat\', fs_opts={\'anon\': True}) as f:\n        for burst in f.read():\n            print(burst.header)\n```\n\n## Install\n\nThe package can be installed from PyPI (note the package distribution name):\n\n```bash\n$ pip install bas-apres\n```\n\n### Install remote extra\n\nTo load data from remote cloud storage platforms, the `remote` extra must be installed:\n\n```bash\n$ pip install bas-apres[remote]\n```\n\nIn addition, specific `fsspec` packages for each cloud storage platform to be used must also be installed.  For example, to load data from Amazon S3 buckets:\n\n```bash\n$ pip install s3fs\n```\n\n## Simple utility scripts\n\nThe package contains a number of scripts for converting from/to netCDF, plotting the data etc.  When installing the package (e.g., via `pip install`) then these scripts are available as commands and added to the `PATH`.  So for instance, you can just call `apres_to_nc`.  If running the scripts from a `clone` of the source code repository, then they need to be run as modules, e.g. `python -m apres.apres_to_nc`.\n\n### Input files held on remote cloud storage\n\nThe utility scripts have the ability to load data from remote cloud storage.  To do so, the input filename must be specified as a valid URI for the particular cloud storage platform.  In addtion, any `fsspec` options required for accessing the remote resource should be specified as a JSON string (via the `-o` option) or as a JSON file (via the `-O` option).  The latter is a better choice if the `fsspec` options include sensitive credentials.  For example, to anonymously access publicly open data from an S3 bucket:\n\n```bash\n$ python -m apres.read_apres -o \'{"anon": true}\' s3://apres-tests/short-test-data.dat\n```\n\nor equivalently, given a JSON file:\n\n```bash\n$ cat fs_opts.json \n{"anon": true}\n$ python -m apres.read_apres -O fs_opts.json s3://apres-tests/short-test-data.dat\n```\n\n### apres_to_nc.py\n\nThis script converts an ApRES .dat file to netCDF4.  The default output netCDF filename has the same name as the input .dat file, but with a .nc file suffix.  Optionally an alternative output netCDF filename can be explicitly given.\n\nThe conversion is a straightforward mapping, in that each burst becomes a netCDF4 group, where the header lines become group attributes, and the data are stored as the group data.  How the data are stored depends on the number of attenuators used when acquiring the data:\n\n* Single-attenuator configuration: The data are stored as a `NSubBursts` x `N_ADC_SAMPLES` 2D data array.\n* Multi-attenuator configuration: The data are stored as a `NSubBursts` x `N_ADC_SAMPLES` x `nAttenuators` 3D data array.\n\n### nc_to_apres.py\n\nThis script converts a previously converted netCDF4 file back to the original ApRES .dat file.  The default output ApRES .dat filename has the same name as the input netCDF file, but with a .dat file suffix.  Optionally an alternative output ApRES filename can be explicitly given.\n\nThe conversion is a straightforward reversal of the original conversion.  For the newer ApRES .dat file format version, this should be identical to the original file.  For the older ApRES .dat file format version, there will likely be small differences in the whitespace in the textual header.  Ignoring these insignificant changes in whitespace (e.g. `diff -awB orig.dat reconstructed.dat`), the files will be identical.\n\n### plot_apres.py\n\nThis script will plot the `N_ADC_SAMPLES` vs `NSubBursts` as a radargram, for each burst in the file (or group for converted netCDF files).  If the `Average` header item > 0 (and so each subburst has been aggregated), then the script will instead plot the first subburst as a single trace.  If the `nAttenuators` header item > 1, then each attenuator\'s data are plotted separately.  The file can be either an ApRES .dat file, or a converted netCDF file.\n\n### read_apres.py\n\nThis script will read the given ApRES .dat file, and for each burst, print the results of parsing the header, such as the dimensions of the data array, and the parsed header dictionary.  It will also *head* the data section (by default the first 10 samples of the first 10 subbursts), to give an initial simple look at the data.  If the data were acquired using multiple attenuators, then the number of samples shown will be multiplied by the number of attenuators.\n\nThe script\'s primary purpose is as a simple example of how to use the `ApRESFile` class to read an ApRES .dat file.\n\n### write_apres.py\n\nThis script will read the given input ApRES .dat file, and for each burst, write the header and data to the given output ApRES .dat file.  Optionally a subset of bursts can be written out, specified as the first `bursts` bursts of the input file.  In addition a subset of each burst can be written out, specified as the first `subbursts` subbursts, and the first `samples` ADC samples of these subbursts.  If `bursts`, `subbursts` and `samples` are not specified, then the output file is identical to the input file.\n\nThe script\'s primary purpose is as a simple example of how to use the `ApRESFile` class to rewrite an ApRES .dat file.\n\n## Command line usage\n\n### apres_to_nc.py\n\n```python\nusage: apres_to_nc [-h] [-o FS_OPTS | -O FS_OPTS] [-V] infile [outfile]\n\nconvert a raw ApRES file to a netCDF4 file\n\npositional arguments:\n  infile                ApRES raw file\n  outfile               converted netCDF file\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -o FS_OPTS, --fs-opts FS_OPTS\n                        fsspec filesystem options from a JSON string\n  -O FS_OPTS, --fs-opts-file FS_OPTS\n                        fsspec filesystem options from a JSON file\n  -V, --version         show program\'s version number and exit\n```\n\n### nc_to_apres.py\n\n```python\nusage: nc_to_apres [-h] [-V] infile [outfile]\n\nrecover the raw ApRES file from a converted netCDF4 file\n\npositional arguments:\n  infile         converted netCDF file of raw ApRES data\n  outfile        ApRES raw file\n\noptional arguments:\n  -h, --help     show this help message and exit\n  -V, --version  show program\'s version number and exit\n```\n\n### plot_apres.py\n\n```python\nusage: plot_apres [-h] [-r | -t] [-g GRID GRID] [-c CONTRAST] [-m CMAP]\n                  [-o FS_OPTS | -O FS_OPTS] [-V]\n                  filename\n\nplot ApRES data, either from a .dat file, or a converted netCDF file\n\npositional arguments:\n  filename              filename\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -r, --radargram       plot a radargram\n  -t, --traces          plot individual traces\n  -g GRID GRID, --grid GRID GRID\n                        plot the first nrows x ncols traces\n  -c CONTRAST, --contrast CONTRAST\n                        contrast for the radargram\n  -m CMAP, --cmap CMAP  colour map for the radargram\n  -o FS_OPTS, --fs-opts FS_OPTS\n                        fsspec filesystem options from a JSON string\n  -O FS_OPTS, --fs-opts-file FS_OPTS\n                        fsspec filesystem options from a JSON file\n  -V, --version         show program\'s version number and exit\n\nExamples\n\nPlot the given ApRES .dat file as a radargram:\n\npython3 -m apres.plot_apres filename.dat\n\nPlot the given converted netCDF file as a radargram:\n\npython3 -m apres.plot_apres filename.nc\n\nPlot the given ApRES file as a radargram, increasing the contrast:\n\npython3 -m apres.plot_apres -c 10 filename.dat\n\nSame as above, but with a blue-white-red colour map:\n\npython3 -m apres.plot_apres -c 10 -m bwr filename.dat\n\nPlot the first trace from the given ApRES file:\n\npython3 -m apres.plot_apres -t filename.dat\n\nPlot the first 6 traces, in a 3x2 grid:\n\npython3 -m apres.plot_apres -t -g 3 2 filename.dat\n```\n\n### read_apres.py\n\n```python\nusage: read_apres [-h] [-o FS_OPTS | -O FS_OPTS] [-V] infile\n\nprint the header, a sample of the data, and diagnostics, from the given raw\nApRES file\n\npositional arguments:\n  infile                ApRES raw file\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -o FS_OPTS, --fs-opts FS_OPTS\n                        fsspec filesystem options from a JSON string\n  -O FS_OPTS, --fs-opts-file FS_OPTS\n                        fsspec filesystem options from a JSON file\n  -V, --version         show program\'s version number and exit\n```\n\n### write_apres.py\n\n```python\nusage: write_apres [-h] [-b BURSTS] [-u SUBBURSTS] [-s SAMPLES]\n                   [-o FS_OPTS | -O FS_OPTS] [-V]\n                   infile outfile\n\nread a raw ApRES file and write to another raw ApRES file, optionally\nsubsetting the data\n\npositional arguments:\n  infile                ApRES raw file\n  outfile               ApRES raw file\n\noptional arguments:\n  -h, --help            show this help message and exit\n  -b BURSTS, --bursts BURSTS\n                        only write out the first BURSTS bursts\n  -u SUBBURSTS, --subbursts SUBBURSTS\n                        only write out the first SUBBURSTS subbursts\n  -s SAMPLES, --samples SAMPLES\n                        only write out the first SAMPLES ADC samples\n  -o FS_OPTS, --fs-opts FS_OPTS\n                        fsspec filesystem options from a JSON string\n  -O FS_OPTS, --fs-opts-file FS_OPTS\n                        fsspec filesystem options from a JSON file\n  -V, --version         show program\'s version number and exit\n```\n',
    'author': 'Paul Breen',
    'author_email': 'pbree@bas.ac.uk',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/antarctica/bas-apres',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)
